================================================================================

The files included in this archive contain the "Leaky Frontend" attack demo code regarding the security 
vulnerabilities of processor frontends.
Please contact us with any questions:

Shuwen Deng (shuwen.deng@yale.edu)
Bowen Huang (bowen.huang@yale.edu)
Jakub Szefer (jakub.szefer@yale.edu)

Code updated: June 1st, 2022

README updated: June 1st, 2022

================================================================================

This code is distributed under GPLv3 license.  Please see LICENSE file for
the details of the license.

================================================================================

System requirements: 

The code is tested on Ubuntu 18.04 with gcc 7.5.0. But it should work on other
linux distributions.

If you want to run on Windows, you need to replace some functions, and
install shared library properly. Good luck!

To test hyper-threaded sharing, you must use a machine with a processor that
has SMT feature enabled.

================================================================================

Step 1: Preparations

--------------------------------------------------------------------------------
Step 1.0: install python2.7 and related libraries including numpy, pandas, and 
distance from scipy.spatial

--------------------------------------------------------------------------------
Step 1.2: check hyper-threaded sharing of your processor:

$ lscpu --extend

And find one *CPU* ids that share the same *Physical Core* as hardware thread 2. 
Here we name it as $SEC_HW_THREAD.

For example, on Intel(R) Xeon(R) E-2286G CPU with 12 cores, CPU 2 and 8 share the 
same physical core.

For each of the following attacks, in the current version, you need to manually
edit the following line by changing the hardcoded core number to $SEC_HW_THREAD:
"CPU_SET(8,&attacker_set);".

--------------------------------------------------------------------------------
Step 1.3: Prepare SGX environment:

You can follow the following link to install and set up SGX environment in your 
machine if the hardware is equipped with SGX:
https://github.com/intel/linux-sgx

Be sure to source the SGX installation environment:
$ source ${sgx-sdk-install-path}/environment

"sgx-gdb" can be used to identify the sender/receiver address in multi-threading
attacks, which helps tune the addresses to map to the same micro-op set.

================================================================================

Step 2: Covert channel of multi-threading (MT) eviction/misalignment-based attack

--------------------------------------------------------------------------------

Step 2.1: Make Intel’s Running Average Power Limit (RAPL) if measuring the power:

$ sudo chmod 755 /sys/devices/virtual/powercap/intel-rapl/intel-rapl\:0/energy_uj

--------------------------------------------------------------------------------

Step 2.2: Run the covert channel

$ cd non_sgx_code/mt_eviction(misalignment)_based_attack
$ ./collect_perf_num_two.sh $OUTPUTFILE $TOTAL_VISIT_DVALUE $SEC_HW_THREAD 
$INPUT_BIT_TYPE $REPEAT_TIME

The format of $OUTPUTFILE can be: asse_loop_data_v$VERSION_input_$INPUT_BIT_TYPE.txt.
$VERSION is the version number of the experiments, e.g., 1.

$INPUT_BIT_TYPE represents different input types.
*0*: all 0s message; *1*: all 1s message; *a*: alternating 0s and 1s message; *r*: 
random message.

$TOTAL_VISIT_DVALUE: for this value we set it to be *8* to try all different *d* 
numbers from 1 to 8.

$SEC_HW_THREAD is the core id number derived from Step 1.2.

$REPEAT_TIME: the total iterated number of experiments for each input type and 
with different d value.

Tips:
* For input type *a* and *r*, suggest to set $REPEAT_TIME to be larger than or equal 
to *60* to avoid influence of insuccessful runs. 
* For input type *0* and *1*, set $REPEAT_TIME to be *20* is enough to have ample 
insuccessful runs. 

Sample running instruction:

$ ./collect_perf_num_two.sh asse_loop_data_v1_input_a.txt 8 8 a 60

This is for running MT Eviction-Based Attack for different values of 
parameter d from 1 to 8 with input type to be alternating 0s and 1s. 

You will expect to get output similar to the non_sgx_code/
mt_eviction_based_attack/asse_loop_data_v1_input_$INPUT_BIT_TYPE.txt.
non_sgx_code/
mt_eviction_based_attack/asse_loop_data_v1_input_$INPUT_BIT_TYPE_message.txt
files are generated in the middle to assist processing the output results shown
in Step 2.3.
--------------------------------------------------------------------------------

Step 2.3: Process the output results

$ cd ../../
$ python process_covert_channel.py $PATH_TO_FILE $REPEAT_TIME $TOTAL_VISIT_DVALUE
$LOOP_NUMBER $INT_INPUT_BIT_TYPE
non_sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_0.txt 60 8 2000 0

$PATH_TO_FILE is the path of the output file derived from Step 2.2, e.g., 
non_sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_0.txt

$REPEAT_TIME: the total iterated number of experiments for each input type and 
with different d value.

$TOTAL_VISIT_DVALUE: for this value we set it to be *8* to try all different *d* 
numbers from 1 to 8.

$LOOP_NUMBER is the total looping number of one MT eviction-based attack，
e.g., 2000.

$INT_INPUT_BIT_TYPE represents different input types using numbers.
*0*: all 0s message; *1*: all 1s message; *2*: alternating 0s and 1s message; *3*: 
random message.


Sample running instruction:

python process_covert_channel.py 
non_sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_0.txt 60 
8 2000 0

This is for processing the results of MT Eviction-Based Attack for 
different values of parameter d from 1 to 8 with input type to be all 0s message.

You will expect to get output similar to the non_sgx_code/
mt_eviction_based_attack/asse_loop_data_v1_input_$INPUT_BIT_TYPE_output.txt.

--------------------------------------------------------------------------------

Step 2.4: Generate the transmission rates and error rates

$ python draw_all_error_rate.py $PATH_TO_FILE $INT_INPUT_BIT_TYPE $NUM_D

$PATH_TO_FILE is the path of the output file derived from Step 2.2, e.g., 
non_sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_0.txt

$INT_INPUT_BIT_TYPE represents different input types using numbers.
*0*: all 0s message; *1*: all 1s message; *2*: alternating 0s and 1s message; *3*: 
random message.

$NUM_D: number of tested different d values. For the given non-MT code, 
eviction-based attack number is less than *8* and misalignment-based attack
is less than *7*.

Sample running instruction:

python draw_all_error_rate.py 
non_sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_r.txt r
This is for generating the transmission rates and error rates of MT 
Eviction-Based Attack for different values of parameter d from 1 to 8 with 
input type to be random message. 

You will expect to get output similar to the non_sgx_code/
mt_eviction_based_attack/asse_loop_data_v1_input_$INPUT_BIT_TYPE_res.txt.




================================================================================

Step 3: Covert channel of non multi-threading (non-MT) stealthy/fast
eviction/misalignment-based attack

The default non-MT covert channel only tests *a* type input bit type. The code 
can be changed to take other input bit types as well.
The default non-MT covert channel takes d_value as 6 (5 for misalignment-based 
attacks) for the optimal cases. The code can be changed to test other d_value 
as well.
The core running the code is by default set to be *3*rd core. It can be changed
in collect_perf_num_single.sh file as well.

--------------------------------------------------------------------------------

Step 3.1: Run the covert channel

$ cd non_sgx_code/nonmt_stealthy(fast)_eviction(misalignment)_based_attack
$ ./collect_perf_num_single.sh $OUTPUTFILE $REPEAT_TIME $LOOP_NUMBER

Parameters follow what have been defined in Step 2. 

Tips:
* Suggest to set $REPEAT_TIME to be larger than or equal to *30* to avoid influence 
of insuccessful runs. 
* Suggest to set $LOOP_NUMBER to be larger than or equal to *250* to maintain
acceptable success rate. When changing the $LOOP_NUMBER, range of decoded numbers
from the covert channel (line 148 in 1T_process_covert_channel.py) also needs to 
be changed accordingly.

Sample running instruction:

$ ./collect_perf_num_single.sh asse_loop_data_v1_input_a.txt 30 250

This is for running non-MT Stealthy Eviction-Based Attack with input type to be 
alternating 0s and 1s. 

You will expect to get output similar to the non_sgx_code/
nonmt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a.txt.

--------------------------------------------------------------------------------

Step 3.2: Process the output results

$ cd ../../
$ python2.7 1T_process_covert_channel.py $PATH_TO_FILE $REPEAT_TIME $NUM_D
$LOOP_NUMBER

$PATH_TO_FILE is the path of the output file derived from Step 3.1, e.g., 
non_sgx_code/nonmt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a.txt

$NUM_D: number of tested different d values. For the given non-MT code, this 
parameter is set to be 1.

Sample running instruction:

python2.7 1T_process_covert_channel.py 
non_sgx_code/nonmt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a.txt 
30 1 250

This is for processing the results of non-MT Stealthy Eviction-Based Attack with 
input type to be alternating 0s and 1s. 

You will expect to get output similar to the non_sgx_code/
nonmt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a_output.txt.

--------------------------------------------------------------------------------

Step 3.3: Generate the transmission rates and error rates

$ python draw_all_error_rate.py $PATH_TO_FILE $INT_INPUT_BIT_TYPE $NUM_D

$INT_INPUT_BIT_TYPE: For the given non-MT code, this parameter is set to be *a*.
$NUM_D: For the given non-MT code, this parameter is set to be 1.

Sample running instruction:

python draw_all_error_rate.py 
non_sgx_code/nonmt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a.txt a 1
This is for generating the transmission rates and error rates of non-MT Stealthy 
Eviction-Based Attack with input type to be alternating 0s and 1s. 

You will expect to get output similar to the non_sgx_code/
mt_stealthy_eviction_based_attack/asse_loop_data_v1_input_a_res.txt.



================================================================================

Step 4: Run covert channel of multi-threading (MT) eviction/misalignment-based 
attack on SGX


The default SGX-version MT covert channel takes d_value as 6 (5 for 
misalignment-based attacks) for the optimal cases. The code can be changed to 
test other d_value as well.

--------------------------------------------------------------------------------

Step 4.1: Run the covert channel

$ cd sgx_code/mt_eviction(misalignment)_based_attack
$ ./collect_perf_num_two.sh $OUTPUTFILE $SEC_HW_THREAD $INPUT_BIT_TYPE $REPEAT_TIME 

Parameters follow what have been defined in Step 2. 

Tips:
* Suggest to set $REPEAT_TIME to be larger than or equal to *60* to avoid influence 
of insuccessful runs. 
* $LOOP_NUMBER is by default set to be *1200* to maintain acceptable success rate.
It can be changed in collect_perf_num_two.sh file. When changing the $LOOP_NUMBER, 
range of decoded numbers from the covert channel (line 153-156 in 
process_covert_channel.py) also needs to be changed accordingly.

Sample running instruction:

$ ./collect_perf_num_two.sh asse_loop_data_v1_input_a.txt 8 a 60

This is for running SGX-version MT Eviction-Based Attack with input type to be 
alternating 0s and 1s. 

You will expect to get output similar to the sgx_code/
mt_eviction_based_attack/asse_loop_data_v1_input_a.txt.

--------------------------------------------------------------------------------

Step 4.2: Process the output results

$ cd ../../
$ python2.7 process_covert_channel.py $PATH_TO_FILE $REPEAT_TIME $NUM_D
$LOOP_NUMBER $INPUT_BIT_TYPE 

$PATH_TO_FILE is the path of the output file derived from Step 4.1, e.g., 
sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_a.txt

$NUM_D: number of tested different d values. For the given MT code, this 
parameter is set to be 1.

Sample running instruction:

python2.7 process_covert_channel.py 
sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_a.txt 60 1 1200 2

This is for processing the results of SGX-version MT Eviction-Based Attack 
with input type to be alternating 0s and 1s. 

You will expect to get output similar to the 
sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_a_output.txt

--------------------------------------------------------------------------------

Step 4.3: Generate the transmission rates and error rates

$ python draw_all_error_rate.py $PATH_TO_FILE $INT_INPUT_BIT_TYPE $NUM_D

$NUM_D: For the given MT code, this parameter is set to be 1.

Sample running instruction:

python draw_all_error_rate.py 
sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_a.txt a 1
This is for generating the transmission rates and error rates of SGX-version MT 
Eviction-Based Attack with input type to be alternating 0s and 1s. 

You will expect to get output similar to the 
sgx_code/mt_eviction_based_attack/asse_loop_data_v1_input_a_res.txt.



================================================================================

Step 5: Run covert channel of non multi-threading (non-MT) stealthy/fast
eviction/misalignment-based attack on SGX

The default non-MT covert channel only tests *a* type input bit type. The code 
can be changed to take other input bit types as well.
The default non-MT covert channel takes d_value as 6 (5 for misalignment-based 
attacks) for the optimal cases. The code can be changed to test other d_value 
as well.
Not mentioned parameter definitions follow what have been defined in Step 3.

--------------------------------------------------------------------------------

Step 5.1: Run the covert channel

$ cd sgx_code/nonmt_stealthy(fast)_eviction(misalignment)_based_attack
$ ./collect_perf_num_single.sh $OUTPUTFILE $REPEAT_TIME $LOOP_NUMBER

Parameters follow what have been defined in Step 2. 

Tips:
* Suggest to set $REPEAT_TIME to be larger than or equal to *60* to avoid influence 
of insuccessful runs. 
* Suggest to set $LOOP_NUMBER to be larger than or equal to *250* to maintain
acceptable success rate.

Sample running instruction:

$ ./collect_perf_num_single.sh asse_loop_data_v1_input_a.txt 60 250

This is for running SGX-version non-MT Fast Misalignment-Based Attack with input 
type to be alternating 0s and 1s. 

You will expect to get output similar to the sgx_code/
nonmt_fast_misalignment_based_attack/asse_loop_data_v1_input_a.txt.

--------------------------------------------------------------------------------

Step 5.2: Process the output results

$ cd ../../
$ python2 1T_process_covert_channel.py $PATH_TO_FILE $REPEAT_TIME $NUM_D
$LOOP_NUMBER

$PATH_TO_FILE is the path of the output file derived from Step 5.1, e.g., 
sgx_code/nonmt_fast_misalignment_based_attack/asse_loop_data_v1_input_a.txt

$NUM_D: number of tested different d values. For the given non-MT code, this 
parameter is set to be 1.

Sample running instruction:

python2 1T_process_covert_channel.py sgx_code/nonmt_fast_misalignment_based_attack/asse_loop_data_v1_input_a.txt 60 1 250

This is for processing the results of SGX-version non-MT Fast Misalignment-Based 
Attack with input type to be alternating 0s and 1s. 

You will expect to get output similar to the 
sgx_code/nonmt_fast_misalignment_based_attack/asse_loop_data_v1_input_a_output.txt

--------------------------------------------------------------------------------

Step 5.3: Generate the transmission rates and error rates

$ python draw_all_error_rate.py $PATH_TO_FILE $INT_INPUT_BIT_TYPE $NUM_D

$INT_INPUT_BIT_TYPE: For the given non-MT code, this parameter is set to be *a*.
$NUM_D: For the given non-MT code, this parameter is set to be 1.

Sample running instruction:

python draw_all_error_rate.py 
sgx_code/nonmt_fast_misaligment_based_attack/asse_loop_data_v1_input_a.txt a 1
This is for generating the transmission rates and error rates of SGX-version non-MT 
Fast Misalignment-Based Attack with input type to be alternating 0s and 1s. 

You will expect to get output similar to the 
sgx_code/nonmt_fast_misaligment_based_attack/asse_loop_data_v1_input_a_res.txt
