#include "encl_t.h"

#include "sgx_trts.h" /* for sgx_ocalloc, sgx_is_outside_enclave */
#include "sgx_lfence.h" /* for sgx_lfence */

#include <errno.h>
#include <mbusafecrt.h> /* for memcpy_s etc */
#include <stdlib.h> /* for malloc/free etc */

#define CHECK_REF_POINTER(ptr, siz) do {	\
	if (!(ptr) || ! sgx_is_outside_enclave((ptr), (siz)))	\
		return SGX_ERROR_INVALID_PARAMETER;\
} while (0)

#define CHECK_UNIQUE_POINTER(ptr, siz) do {	\
	if ((ptr) && ! sgx_is_outside_enclave((ptr), (siz)))	\
		return SGX_ERROR_INVALID_PARAMETER;\
} while (0)

#define CHECK_ENCLAVE_POINTER(ptr, siz) do {	\
	if ((ptr) && ! sgx_is_within_enclave((ptr), (siz)))	\
		return SGX_ERROR_INVALID_PARAMETER;\
} while (0)

#define ADD_ASSIGN_OVERFLOW(a, b) (	\
	((a) += (b)) < (b)	\
)


static sgx_status_t SGX_CDECL sgx_ecall_secret_lookup(void* pms)
{
	sgx_status_t status = SGX_SUCCESS;

#ifdef _MSC_VER
	/* In case enclave `encl' doesn't call any tRTS function. */
	volatile int force_link_trts = sgx_is_within_enclave(NULL, 0);
	(void) force_link_trts; /* avoid compiler warning */
#endif

	if (pms != NULL) return SGX_ERROR_INVALID_PARAMETER;
	ecall_secret_lookup();
	return status;
}

static sgx_status_t SGX_CDECL sgx_ecall_secret_lookup_notmap(void* pms)
{
	sgx_status_t status = SGX_SUCCESS;

#ifdef _MSC_VER
	/* In case enclave `encl' doesn't call any tRTS function. */
	volatile int force_link_trts = sgx_is_within_enclave(NULL, 0);
	(void) force_link_trts; /* avoid compiler warning */
#endif

	if (pms != NULL) return SGX_ERROR_INVALID_PARAMETER;
	ecall_secret_lookup_notmap();
	return status;
}

SGX_EXTERNC const struct {
	size_t nr_ecall;
	struct {void* ecall_addr; uint8_t is_priv; uint8_t is_switchless;} ecall_table[2];
} g_ecall_table = {
	2,
	{
		{(void*)(uintptr_t)sgx_ecall_secret_lookup, 0, 0},
		{(void*)(uintptr_t)sgx_ecall_secret_lookup_notmap, 0, 0},
	}
};

SGX_EXTERNC const struct {
	size_t nr_ocall;
} g_dyn_entry_table = {
	0,
};


